@@ -1,13 +1,21 @@
 from __future__ import annotations
 
 import functools
+import operator
 import re
-from contextlib import suppress
-from typing import TYPE_CHECKING, Any, Callable, Literal, Sized, TypeVar
+from typing import TYPE_CHECKING, Any, Callable, Literal, TypeVar
 
 import pandas as pd
 
-from narwhals._compliant.series import EagerSeriesNamespace
+from narwhals._compliant import EagerSeriesNamespace
+from narwhals._constants import (
+    MS_PER_SECOND,
+    NS_PER_MICROSECOND,
+    NS_PER_MILLISECOND,
+    NS_PER_SECOND,
+    SECONDS_PER_DAY,
+    US_PER_SECOND,
+)
 from narwhals._utils import (
     Implementation,
     Version,
@@ -15,19 +23,32 @@ from narwhals._utils import (
     check_columns_exist,
     isinstance_or_issubclass,
 )
-from narwhals.exceptions import DuplicateError, ShapeError
-
-T = TypeVar("T", bound=Sized)
+from narwhals.exceptions import ShapeError
 
 if TYPE_CHECKING:
-    from pandas._typing import Dtype as PandasDtype
+    from collections.abc import Iterable, Iterator, Mapping
+    from types import ModuleType
 
+    from pandas._typing import Dtype as PandasDtype
+    from pandas.core.dtypes.dtypes import BaseMaskedDtype
+    from typing_extensions import TypeAlias, TypeIs
+
+    from narwhals._duration import IntervalUnit
     from narwhals._pandas_like.expr import PandasLikeExpr
     from narwhals._pandas_like.series import PandasLikeSeries
+    from narwhals._pandas_like.typing import (
+        NativeDataFrameT,
+        NativeNDFrameT,
+        NativeSeriesT,
+    )
     from narwhals.dtypes import DType
-    from narwhals.typing import DTypeBackend, TimeUnit, _1DArray
+    from narwhals.typing import DTypeBackend, IntoDType, TimeUnit, _1DArray
 
     ExprT = TypeVar("ExprT", bound=PandasLikeExpr)
+    UnitCurrent: TypeAlias = TimeUnit
+    UnitTarget: TypeAlias = TimeUnit
+    BinOpBroadcast: TypeAlias = Callable[[Any, int], Any]
+    IntoRhs: TypeAlias = int
 
 
 PANDAS_LIKE_IMPLEMENTATION = {
@@ -81,7 +102,42 @@ PA_DURATION_RGX = r"""^
 $"""
 PATTERN_PA_DURATION = re.compile(PA_DURATION_RGX, re.VERBOSE)
 
-UNIT_DICT = {"d": "D", "m": "min"}
+NativeIntervalUnit: TypeAlias = Literal[
+    "year",
+    "quarter",
+    "month",
+    "week",
+    "day",
+    "hour",
+    "minute",
+    "second",
+    "millisecond",
+    "microsecond",
+    "nanosecond",
+]
+ALIAS_DICT = {"d": "D", "m": "min"}
+UNITS_DICT: Mapping[IntervalUnit, NativeIntervalUnit] = {
+    "y": "year",
+    "q": "quarter",
+    "mo": "month",
+    "d": "day",
+    "h": "hour",
+    "m": "minute",
+    "s": "second",
+    "ms": "millisecond",
+    "us": "microsecond",
+    "ns": "nanosecond",
+}
+
+PANDAS_VERSION = Implementation.PANDAS._backend_version()
+"""Static backend version for `pandas`.
+
+Always available if we reached here, due to a module-level import.
+"""
+
+
+def is_pandas_or_modin(implementation: Implementation) -> bool:
+    return implementation in {Implementation.PANDAS, Implementation.MODIN}
 
 
 def align_and_extract_native(
@@ -92,14 +148,10 @@ def align_and_extract_native(
     If the comparison isn't supported, return `NotImplemented` so that the
     "right-hand-side" operation (e.g. `__radd__`) can be tried.
     """
-    from narwhals._pandas_like.dataframe import PandasLikeDataFrame
     from narwhals._pandas_like.series import PandasLikeSeries
 
     lhs_index = lhs.native.index
 
-    if isinstance(rhs, PandasLikeDataFrame):
-        return NotImplemented
-
     if lhs._broadcast and isinstance(rhs, PandasLikeSeries) and not rhs._broadcast:
         return lhs.native.iloc[0], rhs.native
 
@@ -109,12 +161,7 @@ def align_and_extract_native(
         if rhs.native.index is not lhs_index:
             return (
                 lhs.native,
-                set_index(
-                    rhs.native,
-                    lhs_index,
-                    implementation=rhs._implementation,
-                    backend_version=rhs._backend_version,
-                ),
+                set_index(rhs.native, lhs_index, implementation=rhs._implementation),
             )
         return (lhs.native, rhs.native)
 
@@ -126,12 +173,8 @@ def align_and_extract_native(
 
 
 def set_index(
-    obj: T,
-    index: Any,
-    *,
-    implementation: Implementation,
-    backend_version: tuple[int, ...],
-) -> T:
+    obj: NativeNDFrameT, index: Any, *, implementation: Implementation
+) -> NativeNDFrameT:
     """Wrapper around pandas' set_axis to set object index.
 
     We can set `copy` / `inplace` based on implementation/version.
@@ -141,38 +184,26 @@ def set_index(
     ) != (actual_len := len(obj)):
         msg = f"Expected object of length {expected_len}, got length: {actual_len}"
         raise ShapeError(msg)
-    if implementation is Implementation.CUDF:  # pragma: no cover
-        obj = obj.copy(deep=False)  # type: ignore[attr-defined]
-        obj.index = index  # type: ignore[attr-defined]
+    if implementation is Implementation.CUDF:
+        obj = obj.copy(deep=False)
+        obj.index = index
         return obj
     if implementation is Implementation.PANDAS and (
-        backend_version < (1,)
+        (1, 5) <= implementation._backend_version() < (3,)
     ):  # pragma: no cover
-        kwargs = {"inplace": False}
-    else:
-        kwargs = {}
-    if implementation is Implementation.PANDAS and (
-        (1, 5) <= backend_version < (3,)
-    ):  # pragma: no cover
-        kwargs["copy"] = False
-    else:  # pragma: no cover
-        pass
-    return obj.set_axis(index, axis=0, **kwargs)  # type: ignore[attr-defined]
+        return obj.set_axis(index, axis=0, copy=False)
+    return obj.set_axis(index, axis=0)  # pragma: no cover
 
 
 def rename(
-    obj: T,
-    *args: Any,
-    implementation: Implementation,
-    backend_version: tuple[int, ...],
-    **kwargs: Any,
-) -> T:
+    obj: NativeNDFrameT, *args: Any, implementation: Implementation, **kwargs: Any
+) -> NativeNDFrameT:
     """Wrapper around pandas' rename so that we can set `copy` based on implementation/version."""
     if implementation is Implementation.PANDAS and (
-        backend_version >= (3,)
+        implementation._backend_version() >= (3,)
     ):  # pragma: no cover
-        return obj.rename(*args, **kwargs)  # type: ignore[attr-defined]
-    return obj.rename(*args, **kwargs, copy=False)  # type: ignore[attr-defined]
+        return obj.rename(*args, **kwargs, inplace=False)
+    return obj.rename(*args, **kwargs, copy=False, inplace=False)
 
 
 @functools.lru_cache(maxsize=16)
@@ -212,7 +243,15 @@ def non_object_native_to_narwhals_dtype(native_dtype: Any, version: Version) ->
         "float[pyarrow]",
     }:
         return dtypes.Float32()
-    if dtype in {"string", "string[python]", "string[pyarrow]", "large_string[pyarrow]"}:
+    if dtype in {
+        # "there is no problem which can't be solved by adding an extra string type" pandas
+        "string",
+        "string[python]",
+        "string[pyarrow]",
+        "string[pyarrow_numpy]",
+        "large_string[pyarrow]",
+        "str",
+    }:
         return dtypes.String()
     if dtype in {"bool", "boolean", "boolean[pyarrow]", "bool[pyarrow]"}:
         return dtypes.Boolean()
@@ -246,7 +285,7 @@ def object_native_to_narwhals_dtype(
     series: PandasLikeSeries, version: Version, implementation: Implementation
 ) -> DType:
     dtypes = version.dtypes
-    if implementation is Implementation.CUDF:  # pragma: no cover
+    if implementation is Implementation.CUDF:
         # Per conversations with their maintainers, they don't support arbitrary
         # objects, so we can just return String.
         return dtypes.String()
@@ -258,7 +297,7 @@ def object_native_to_narwhals_dtype(
     if inferred_dtype == "empty" and version is not Version.V1:
         # Default to String for empty Series.
         return dtypes.String()
-    elif inferred_dtype == "empty":
+    if inferred_dtype == "empty":
         # But preserve returning Object in V1.
         return dtypes.Object()
     return dtypes.Object()
@@ -314,7 +353,7 @@ def native_to_narwhals_dtype(
         )
     if str_dtype != "object":
         return non_object_native_to_narwhals_dtype(native_dtype, version)
-    elif implementation is Implementation.DASK:
+    if implementation is Implementation.DASK:
         # Per conversations with their maintainers, they don't support arbitrary
         # objects, so we can just return String.
         return version.dtypes.String()
@@ -324,6 +363,34 @@ def native_to_narwhals_dtype(
     raise AssertionError(msg)
 
 
+if Implementation.PANDAS._backend_version() >= (1, 2):
+
+    def is_dtype_numpy_nullable(dtype: Any) -> TypeIs[BaseMaskedDtype]:
+        """Return `True` if `dtype` is `"numpy_nullable"`."""
+        # NOTE: We need a sentinel as the positive case is `BaseMaskedDtype.base = None`
+        # See https://github.com/narwhals-dev/narwhals/pull/2740#discussion_r2171667055
+        sentinel = object()
+        return (
+            isinstance(dtype, pd.api.extensions.ExtensionDtype)
+            and getattr(dtype, "base", sentinel) is None
+        )
+else:  # pragma: no cover
+
+    def is_dtype_numpy_nullable(dtype: Any) -> TypeIs[BaseMaskedDtype]:
+        # NOTE: `base` attribute was added between 1.1-1.2
+        # Checking by isinstance requires using an import path that is no longer valid
+        # `1.1`: https://github.com/pandas-dev/pandas/blob/b5958ee1999e9aead1938c0bba2b674378807b3d/pandas/core/arrays/masked.py#L37
+        # `1.2`: https://github.com/pandas-dev/pandas/blob/7c48ff4409c622c582c56a5702373f726de08e96/pandas/core/arrays/masked.py#L41
+        # `1.5`: https://github.com/pandas-dev/pandas/blob/35b0d1dcadf9d60722c055ee37442dc76a29e64c/pandas/core/dtypes/dtypes.py#L1609
+        if isinstance(dtype, pd.api.extensions.ExtensionDtype):
+            from pandas.core.arrays.masked import (  # type: ignore[attr-defined]
+                BaseMaskedDtype as OldBaseMaskedDtype,  # pyright: ignore[reportAttributeAccessIssue]
+            )
+
+            return isinstance(dtype, OldBaseMaskedDtype)
+        return False
+
+
 def get_dtype_backend(dtype: Any, implementation: Implementation) -> DTypeBackend:
     """Get dtype backend for pandas type.
 
@@ -331,117 +398,94 @@ def get_dtype_backend(dtype: Any, implementation: Implementation) -> DTypeBacken
     """
     if implementation is Implementation.CUDF:
         return None
-    if hasattr(pd, "ArrowDtype") and isinstance(dtype, pd.ArrowDtype):
+    if is_dtype_pyarrow(dtype):
         return "pyarrow"
-    with suppress(AttributeError):
-        sentinel = object()
-        if (
-            isinstance(dtype, pd.api.extensions.ExtensionDtype)
-            and getattr(dtype, "base", sentinel) is None
-        ):
-            return "numpy_nullable"
-    return None
+    return "numpy_nullable" if is_dtype_numpy_nullable(dtype) else None
+
+
+# NOTE: Use this to avoid annotating inline
+def iter_dtype_backends(
+    dtypes: Iterable[Any], implementation: Implementation
+) -> Iterator[DTypeBackend]:
+    """Yield a `DTypeBackend` per-dtype.
+
+    Matches pandas' `dtype_backend` argument in `convert_dtypes`.
+    """
+    return (get_dtype_backend(dtype, implementation) for dtype in dtypes)
 
 
 @functools.lru_cache(maxsize=16)
-def is_pyarrow_dtype_backend(dtype: Any, implementation: Implementation) -> bool:
-    return get_dtype_backend(dtype, implementation) == "pyarrow"
+def is_dtype_pyarrow(dtype: Any) -> TypeIs[pd.ArrowDtype]:
+    return hasattr(pd, "ArrowDtype") and isinstance(dtype, pd.ArrowDtype)
 
 
-def narwhals_to_native_dtype(  # noqa: C901, PLR0912, PLR0915
-    dtype: DType | type[DType],
+dtypes = Version.MAIN.dtypes
+NW_TO_PD_DTYPES_INVARIANT: Mapping[type[DType], str] = {
+    # TODO(Unassigned): is there no pyarrow-backed categorical?
+    # or at least, convert_dtypes(dtype_backend='pyarrow') doesn't
+    # convert to it?
+    dtypes.Categorical: "category",
+    dtypes.Object: "object",
+}
+NW_TO_PD_DTYPES_BACKEND: Mapping[type[DType], Mapping[DTypeBackend, str | type[Any]]] = {
+    dtypes.Float64: {
+        "pyarrow": "Float64[pyarrow]",
+        "numpy_nullable": "Float64",
+        None: "float64",
+    },
+    dtypes.Float32: {
+        "pyarrow": "Float32[pyarrow]",
+        "numpy_nullable": "Float32",
+        None: "float32",
+    },
+    dtypes.Int64: {"pyarrow": "Int64[pyarrow]", "numpy_nullable": "Int64", None: "int64"},
+    dtypes.Int32: {"pyarrow": "Int32[pyarrow]", "numpy_nullable": "Int32", None: "int32"},
+    dtypes.Int16: {"pyarrow": "Int16[pyarrow]", "numpy_nullable": "Int16", None: "int16"},
+    dtypes.Int8: {"pyarrow": "Int8[pyarrow]", "numpy_nullable": "Int8", None: "int8"},
+    dtypes.UInt64: {
+        "pyarrow": "UInt64[pyarrow]",
+        "numpy_nullable": "UInt64",
+        None: "uint64",
+    },
+    dtypes.UInt32: {
+        "pyarrow": "UInt32[pyarrow]",
+        "numpy_nullable": "UInt32",
+        None: "uint32",
+    },
+    dtypes.UInt16: {
+        "pyarrow": "UInt16[pyarrow]",
+        "numpy_nullable": "UInt16",
+        None: "uint16",
+    },
+    dtypes.UInt8: {"pyarrow": "UInt8[pyarrow]", "numpy_nullable": "UInt8", None: "uint8"},
+    dtypes.String: {"pyarrow": "string[pyarrow]", "numpy_nullable": "string", None: str},
+    dtypes.Boolean: {
+        "pyarrow": "boolean[pyarrow]",
+        "numpy_nullable": "boolean",
+        None: "bool",
+    },
+}
+UNSUPPORTED_DTYPES = (dtypes.Decimal,)
+
+
+def narwhals_to_native_dtype(  # noqa: C901, PLR0912
+    dtype: IntoDType,
     dtype_backend: DTypeBackend,
     implementation: Implementation,
-    backend_version: tuple[int, ...],
     version: Version,
 ) -> str | PandasDtype:
-    if dtype_backend is not None and dtype_backend not in {"pyarrow", "numpy_nullable"}:
+    if dtype_backend not in {None, "pyarrow", "numpy_nullable"}:
         msg = f"Expected one of {{None, 'pyarrow', 'numpy_nullable'}}, got: '{dtype_backend}'"
         raise ValueError(msg)
     dtypes = version.dtypes
-    if isinstance_or_issubclass(dtype, dtypes.Decimal):
-        msg = "Casting to Decimal is not supported yet."
-        raise NotImplementedError(msg)
-    if isinstance_or_issubclass(dtype, dtypes.Float64):
-        if dtype_backend == "pyarrow":
-            return "Float64[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "Float64"
-        return "float64"
-    if isinstance_or_issubclass(dtype, dtypes.Float32):
-        if dtype_backend == "pyarrow":
-            return "Float32[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "Float32"
-        return "float32"
-    if isinstance_or_issubclass(dtype, dtypes.Int64):
-        if dtype_backend == "pyarrow":
-            return "Int64[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "Int64"
-        return "int64"
-    if isinstance_or_issubclass(dtype, dtypes.Int32):
-        if dtype_backend == "pyarrow":
-            return "Int32[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "Int32"
-        return "int32"
-    if isinstance_or_issubclass(dtype, dtypes.Int16):
-        if dtype_backend == "pyarrow":
-            return "Int16[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "Int16"
-        return "int16"
-    if isinstance_or_issubclass(dtype, dtypes.Int8):
-        if dtype_backend == "pyarrow":
-            return "Int8[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "Int8"
-        return "int8"
-    if isinstance_or_issubclass(dtype, dtypes.UInt64):
-        if dtype_backend == "pyarrow":
-            return "UInt64[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "UInt64"
-        return "uint64"
-    if isinstance_or_issubclass(dtype, dtypes.UInt32):
-        if dtype_backend == "pyarrow":
-            return "UInt32[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "UInt32"
-        return "uint32"
-    if isinstance_or_issubclass(dtype, dtypes.UInt16):
-        if dtype_backend == "pyarrow":
-            return "UInt16[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "UInt16"
-        return "uint16"
-    if isinstance_or_issubclass(dtype, dtypes.UInt8):
-        if dtype_backend == "pyarrow":
-            return "UInt8[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "UInt8"
-        return "uint8"
-    if isinstance_or_issubclass(dtype, dtypes.String):
-        if dtype_backend == "pyarrow":
-            return "string[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "string"
-        return str
-    if isinstance_or_issubclass(dtype, dtypes.Boolean):
-        if dtype_backend == "pyarrow":
-            return "boolean[pyarrow]"
-        elif dtype_backend == "numpy_nullable":
-            return "boolean"
-        return "bool"
-    if isinstance_or_issubclass(dtype, dtypes.Categorical):
-        # TODO(Unassigned): is there no pyarrow-backed categorical?
-        # or at least, convert_dtypes(dtype_backend='pyarrow') doesn't
-        # convert to it?
-        return "category"
+    base_type = dtype.base_type()
+    if pd_type := NW_TO_PD_DTYPES_INVARIANT.get(base_type):
+        return pd_type
+    if into_pd_type := NW_TO_PD_DTYPES_BACKEND.get(base_type):
+        return into_pd_type[dtype_backend]
     if isinstance_or_issubclass(dtype, dtypes.Datetime):
         # Pandas does not support "ms" or "us" time units before version 2.0
-        if implementation is Implementation.PANDAS and backend_version < (
+        if is_pandas_or_modin(implementation) and PANDAS_VERSION < (
             2,
         ):  # pragma: no cover
             dt_time_unit = "ns"
@@ -451,11 +495,10 @@ def narwhals_to_native_dtype(  # noqa: C901, PLR0912, PLR0915
         if dtype_backend == "pyarrow":
             tz_part = f", tz={tz}" if (tz := dtype.time_zone) else ""
             return f"timestamp[{dt_time_unit}{tz_part}][pyarrow]"
-        else:
-            tz_part = f", {tz}" if (tz := dtype.time_zone) else ""
-            return f"datetime64[{dt_time_unit}{tz_part}]"
+        tz_part = f", {tz}" if (tz := dtype.time_zone) else ""
+        return f"datetime64[{dt_time_unit}{tz_part}]"
     if isinstance_or_issubclass(dtype, dtypes.Duration):
-        if implementation is Implementation.PANDAS and backend_version < (
+        if is_pandas_or_modin(implementation) and PANDAS_VERSION < (
             2,
         ):  # pragma: no cover
             du_time_unit = "ns"
@@ -468,9 +511,11 @@ def narwhals_to_native_dtype(  # noqa: C901, PLR0912, PLR0915
         )
     if isinstance_or_issubclass(dtype, dtypes.Date):
         try:
-            import pyarrow as pa  # ignore-banned-import
-        except ModuleNotFoundError:  # pragma: no cover
-            msg = "'pyarrow>=11.0.0' is required for `Date` dtype."
+            import pyarrow as pa  # ignore-banned-import  # noqa: F401
+        except ModuleNotFoundError as exc:  # pragma: no cover
+            # BUG: Never re-raised?
+            msg = "'pyarrow>=13.0.0' is required for `Date` dtype."
+            raise ModuleNotFoundError(msg) from exc
         return "date32[pyarrow]"
     if isinstance_or_issubclass(dtype, dtypes.Enum):
         if version is Version.V1:
@@ -481,72 +526,34 @@ def narwhals_to_native_dtype(  # noqa: C901, PLR0912, PLR0915
             return ns.CategoricalDtype(dtype.categories, ordered=True)
         msg = "Can not cast / initialize Enum without categories present"
         raise ValueError(msg)
-
-    if isinstance_or_issubclass(
-        dtype, (dtypes.Struct, dtypes.Array, dtypes.List, dtypes.Time, dtypes.Binary)
+    if issubclass(
+        base_type, (dtypes.Struct, dtypes.Array, dtypes.List, dtypes.Time, dtypes.Binary)
     ):
-        if implementation is Implementation.PANDAS and backend_version >= (2, 2):
-            try:
-                import pandas as pd
-                import pyarrow as pa  # ignore-banned-import  # noqa: F401
-            except ImportError as exc:  # pragma: no cover
-                msg = f"Unable to convert to {dtype} to to the following exception: {exc.msg}"
-                raise ImportError(msg) from exc
-            from narwhals._arrow.utils import (
-                narwhals_to_native_dtype as arrow_narwhals_to_native_dtype,
-            )
-
-            return pd.ArrowDtype(arrow_narwhals_to_native_dtype(dtype, version=version))
-        else:  # pragma: no cover
-            msg = (
-                f"Converting to {dtype} dtype is not supported for implementation "
-                f"{implementation} and version {version}."
-            )
-            raise NotImplementedError(msg)
+        return narwhals_to_native_arrow_dtype(dtype, implementation, version)
+    if issubclass(base_type, UNSUPPORTED_DTYPES):
+        msg = f"Converting to {base_type.__name__} dtype is not supported for {implementation}."
+        raise NotImplementedError(msg)
     msg = f"Unknown dtype: {dtype}"  # pragma: no cover
     raise AssertionError(msg)
 
 
-def align_series_full_broadcast(*series: PandasLikeSeries) -> list[PandasLikeSeries]:
-    # Ensure all of `series` have the same length and index. Scalars get broadcasted to
-    # the full length of the longest Series. This is useful when you need to construct a
-    # full Series anyway (e.g. `DataFrame.select`). It should not be used in binary operations,
-    # such as `nw.col('a') - nw.col('a').mean()`, because then it's more efficient to extract
-    # the right-hand-side's single element as a scalar.
-    native_namespace = series[0].__native_namespace__()
+def narwhals_to_native_arrow_dtype(
+    dtype: IntoDType, implementation: Implementation, version: Version
+) -> pd.ArrowDtype:
+    if is_pandas_or_modin(implementation) and PANDAS_VERSION >= (2, 2):
+        try:
+            import pyarrow as pa  # ignore-banned-import  # noqa: F401
+        except ImportError as exc:  # pragma: no cover
+            msg = f"Unable to convert to {dtype} to to the following exception: {exc.msg}"
+            raise ImportError(msg) from exc
+        from narwhals._arrow.utils import narwhals_to_native_dtype as _to_arrow_dtype
 
-    lengths = [len(s) for s in series]
-    max_length = max(lengths)
-
-    idx = series[lengths.index(max_length)].native.index
-    reindexed = []
-    for s in series:
-        if s._broadcast:
-            reindexed.append(
-                s._with_native(
-                    native_namespace.Series(
-                        [s.native.iloc[0]] * max_length,
-                        index=idx,
-                        name=s.name,
-                        dtype=s.native.dtype,
-                    )
-                )
-            )
-
-        elif s.native.index is not idx:
-            reindexed.append(
-                s._with_native(
-                    set_index(
-                        s.native,
-                        idx,
-                        implementation=s._implementation,
-                        backend_version=s._backend_version,
-                    )
-                )
-            )
-        else:
-            reindexed.append(s)
-    return reindexed
+        return pd.ArrowDtype(_to_arrow_dtype(dtype, version))
+    msg = (  # pragma: no cover
+        f"Converting to {dtype} dtype is not supported for implementation "
+        f"{implementation} and version {version}."
+    )
+    raise NotImplementedError(msg)
 
 
 def int_dtype_mapper(dtype: Any) -> str:
@@ -557,117 +564,98 @@ def int_dtype_mapper(dtype: Any) -> str:
     return "int64"
 
 
-def calculate_timestamp_datetime(  # noqa: C901, PLR0912
-    s: pd.Series[int], original_time_unit: str, time_unit: str
-) -> pd.Series[int]:
-    if original_time_unit == "ns":
-        if time_unit == "ns":
-            result = s
-        elif time_unit == "us":
-            result = s // 1_000
-        else:
-            result = s // 1_000_000
-    elif original_time_unit == "us":
-        if time_unit == "ns":
-            result = s * 1_000
-        elif time_unit == "us":
-            result = s
-        else:
-            result = s // 1_000
-    elif original_time_unit == "ms":
-        if time_unit == "ns":
-            result = s * 1_000_000
-        elif time_unit == "us":
-            result = s * 1_000
-        else:
-            result = s
-    elif original_time_unit == "s":
-        if time_unit == "ns":
-            result = s * 1_000_000_000
-        elif time_unit == "us":
-            result = s * 1_000_000
-        else:
-            result = s * 1_000
-    else:  # pragma: no cover
-        msg = f"unexpected time unit {original_time_unit}, please report a bug at https://github.com/narwhals-dev/narwhals"
-        raise AssertionError(msg)
-    return result
+_TIMESTAMP_DATETIME_OP_FACTOR: Mapping[
+    tuple[UnitCurrent, UnitTarget], tuple[BinOpBroadcast, IntoRhs]
+] = {
+    ("ns", "us"): (operator.floordiv, 1_000),
+    ("ns", "ms"): (operator.floordiv, 1_000_000),
+    ("us", "ns"): (operator.mul, NS_PER_MICROSECOND),
+    ("us", "ms"): (operator.floordiv, 1_000),
+    ("ms", "ns"): (operator.mul, NS_PER_MILLISECOND),
+    ("ms", "us"): (operator.mul, 1_000),
+    ("s", "ns"): (operator.mul, NS_PER_SECOND),
+    ("s", "us"): (operator.mul, US_PER_SECOND),
+    ("s", "ms"): (operator.mul, MS_PER_SECOND),
+}
 
 
-def calculate_timestamp_date(s: pd.Series[int], time_unit: str) -> pd.Series[int]:
-    s = s * 86_400  # number of seconds in a day
-    if time_unit == "ns":
-        result = s * 1_000_000_000
-    elif time_unit == "us":
-        result = s * 1_000_000
-    else:
-        result = s * 1_000
-    return result
+def calculate_timestamp_datetime(
+    s: NativeSeriesT, current: TimeUnit, time_unit: TimeUnit
+) -> NativeSeriesT:
+    if current == time_unit:
+        return s
+    if item := _TIMESTAMP_DATETIME_OP_FACTOR.get((current, time_unit)):
+        fn, factor = item
+        return fn(s, factor)
+    msg = (  # pragma: no cover
+        f"unexpected time unit {current}, please report an issue at "
+        "https://github.com/narwhals-dev/narwhals"
+    )
+    raise AssertionError(msg)
+
+
+_TIMESTAMP_DATE_FACTOR: Mapping[TimeUnit, int] = {
+    "ns": NS_PER_SECOND,
+    "us": US_PER_SECOND,
+    "ms": MS_PER_SECOND,
+    "s": 1,
+}
+
+
+def calculate_timestamp_date(s: NativeSeriesT, time_unit: TimeUnit) -> NativeSeriesT:
+    return s * SECONDS_PER_DAY * _TIMESTAMP_DATE_FACTOR[time_unit]
 
 
 def select_columns_by_name(
-    df: T,
+    df: NativeDataFrameT,
     column_names: list[str] | _1DArray,  # NOTE: Cannot be a tuple!
-    backend_version: tuple[int, ...],
     implementation: Implementation,
-) -> T:
+) -> NativeDataFrameT | Any:
     """Select columns by name.
 
     Prefer this over `df.loc[:, column_names]` as it's
     generally more performant.
     """
-    if len(column_names) == df.shape[1] and all(column_names == df.columns):  # type: ignore[attr-defined]
+    if len(column_names) == df.shape[1] and (df.columns == column_names).all():
         return df
-    if (df.columns.dtype.kind == "b") or (  # type: ignore[attr-defined]
-        implementation is Implementation.PANDAS and backend_version < (1, 5)
+    if (df.columns.dtype.kind == "b") or (
+        implementation is Implementation.PANDAS
+        and implementation._backend_version() < (1, 5)
     ):
         # See https://github.com/narwhals-dev/narwhals/issues/1349#issuecomment-2470118122
         # for why we need this
-        if error := check_columns_exist(
-            column_names,  # type: ignore[arg-type]
-            available=df.columns.tolist(),  # type: ignore[attr-defined]
-        ):
+        if error := check_columns_exist(column_names, available=df.columns.tolist()):
             raise error
-        return df.loc[:, column_names]  # type: ignore[attr-defined]
+        return df.loc[:, column_names]
     try:
-        return df[column_names]  # type: ignore[index]
+        return df[column_names]
     except KeyError as e:
-        if error := check_columns_exist(
-            column_names,  # type: ignore[arg-type]
-            available=df.columns.tolist(),  # type: ignore[attr-defined]
-        ):
+        if error := check_columns_exist(column_names, available=df.columns.tolist()):
             raise error from e
         raise
 
 
-def check_column_names_are_unique(columns: pd.Index[str]) -> None:
-    try:
-        len_unique_columns = len(columns.drop_duplicates())
-    except Exception:  # noqa: BLE001  # pragma: no cover
-        msg = f"Expected hashable (e.g. str or int) column names, got: {columns}"
-        raise ValueError(msg) from None
-
-    if len(columns) != len_unique_columns:
-        from collections import Counter
-
-        counter = Counter(columns)
-        msg = ""
-        for key, value in counter.items():
-            if value > 1:
-                msg += f"\n- '{key}' {value} times"
-        msg = f"Expected unique column names, got:{msg}"
-        raise DuplicateError(msg)
+def is_non_nullable_boolean(s: PandasLikeSeries) -> bool:
+    # cuDF booleans are nullable but the native dtype is still 'bool'.
+    return (
+        s._implementation
+        in {Implementation.PANDAS, Implementation.MODIN, Implementation.DASK}
+        and s.native.dtype == "bool"
+    )
 
 
-class PandasLikeSeriesNamespace(EagerSeriesNamespace["PandasLikeSeries", Any]):
-    @property
-    def implementation(self) -> Implementation:
-        return self.compliant._implementation
+def import_array_module(implementation: Implementation, /) -> ModuleType:
+    """Returns numpy or cupy module depending on the given implementation."""
+    if implementation in {Implementation.PANDAS, Implementation.MODIN}:
+        import numpy as np
 
-    @property
-    def backend_version(self) -> tuple[int, ...]:
-        return self.compliant._backend_version
+        return np
+    if implementation is Implementation.CUDF:
+        import cupy as cp  # ignore-banned-import  # cuDF dependency.
 
-    @property
-    def version(self) -> Version:
-        return self.compliant._version
+        return cp
+    msg = f"Expected pandas/modin/cudf, got: {implementation}"  # pragma: no cover
+    raise AssertionError(msg)
+
+
+class PandasLikeSeriesNamespace(EagerSeriesNamespace["PandasLikeSeries", Any]): ...
